from dataclasses import Field, dataclass, _MISSING_TYPE, fields
from carica.interface import SerializableType, PrimativeType, primativeTypesTuple, SerializesToDict
from carica.typeChecking import objectIsShallowSerializable, objectIsDeepSerializable, _DeserializedTypeOverrideProxy, _CallableDeserializedTypeOverrideProxy
from carica.carica import BadTypeHandling, BadTypeBehaviour, ErrorHandling, VariableTrace, log
from carica import exceptions
import typing
import traceback
from typing import Any, Dict, List, Mapping, Set, Tuple, Union, cast, TypeVar
# ignoring a warning here because private type _BaseGenericAlias can't be imported right now.
# it is a necessary import to unify over user-defined and special generics.
from typing import _BaseGenericAlias # type: ignore
import inspect

FIELD_TYPE_TYPES = (type, _BaseGenericAlias, TypeVar, _MISSING_TYPE)
FIELD_TYPE_TYPES_UNION = Union[type, _BaseGenericAlias, TypeVar, _MISSING_TYPE]
UNCASTABLE_TYPES = (Any, None)


def _handleTypeCasts(serializedValue: Any, fieldName: str,
                    possibleTypes: Union[FIELD_TYPE_TYPES_UNION, Tuple[FIELD_TYPE_TYPES_UNION]],
                    c_badTypeHandling: BadTypeHandling, _noLog: bool = False):

    def _log(*args, **kwargs):
        if not _noLog:
            log(*args, **kwargs)

    if not isinstance(possibleTypes, tuple):
        possibleTypes = (possibleTypes,)
    
    # Make sure the serialized type matches
    if not isinstance(serializedValue, cast(Tuple[type], possibleTypes)):
        fieldTypeError = f"Expected one of {'/'.join(str(t) for t in possibleTypes)} for field {fieldName}, " \
                        + f"but received serialized type {type(serializedValue).__name__}"

        # Attempt to cast incorrect types - useful, for example, for tuples (TOML only has lists)
        if c_badTypeHandling.behaviour == BadTypeBehaviour.CAST:
            # This is only used if there is only one potential type
            castException: Union[Exception, None] = None

            # Try to cast to each non-Any type
            for potentialType in possibleTypes:
                if potentialType not in UNCASTABLE_TYPES:
                    # We already know that potentialType is not a generic, but we also cannot cast to a TypeVar
                    if isinstance(potentialType, TypeVar):
                        _log(f"Ignoring potential field type {potentialType} - cannot construct TypeVar")
                        continue
                    potentialType = cast(type, potentialType)

                    # Attempt the cast
                    try:
                        newValue = potentialType(serializedValue)
                    # Failed, move onto the next type in the Union
                    except Exception as e:
                        castException = e
                        continue
                    # Cast was successful
                    else:
                        # Log it if required
                        if c_badTypeHandling.logSuccessfulCast:
                            _log(f"[WARNING] Successfully casted unexpected type for field {fieldName} from type " \
                                + f"{type(serializedValue).__name__} to {type(newValue).__name__}")
                        return newValue

            # No cast was successful.
            # Add exception trace if we have exactly one
            if castException is not None and c_badTypeHandling.includeExceptionTrace and len(possibleTypes) == 1:
                trace = traceback.format_exception(type(castException), castException, castException.__traceback__)
                fieldTypeError += f". Exception: {trace}"
            
            # If Any is allowed, just return the value without erroring
            if Any in possibleTypes:
                return serializedValue
            # Nothing to do if keeping failed field casts, except log if configured to
            elif c_badTypeHandling.keepFailedCast:
                if c_badTypeHandling.logTypeKeeping:
                    _log(f"[WARNING] Keeping original value for mistyped field following failed " \
                        + f"cast. {fieldTypeError}")
                return serializedValue
            # Throw errors if set to reject failed casts
            else:
                errMsg = f"Casting failed for unexpected type. {fieldTypeError}"
                if c_badTypeHandling.rejectType == ErrorHandling.RAISE:
                    raise TypeError(errMsg)
                elif c_badTypeHandling.rejectType == ErrorHandling.LOG:
                    _log(f"[WARNING] {errMsg}")

        # mismatched type behaviour is not CAST. Do nothing for KEEP except log if required
        elif c_badTypeHandling.behaviour == BadTypeBehaviour.KEEP:
            if c_badTypeHandling.logTypeKeeping:
                _log(f"[WARNING] Keeping mistyped field value: {fieldTypeError}")
            return serializedValue

        # Throw errors etc for REJECT handling
        elif c_badTypeHandling.behaviour == BadTypeBehaviour.REJECT:
            if c_badTypeHandling.rejectType == ErrorHandling.RAISE:
                raise TypeError(fieldTypeError)
            elif c_badTypeHandling.rejectType == ErrorHandling.LOG:
                _log(fieldTypeError)
            elif c_badTypeHandling.rejectType == ErrorHandling.IGNORE:
                pass
    
    # serialized value matches the expected types, do nothing
    else:
        return serializedValue


def _deserializeField(fieldName: str, fieldType: Union[type, _BaseGenericAlias, TypeVar, _MISSING_TYPE],
                        serializedValue: PrimativeType, c_variableTrace: VariableTrace = [],
                        c_badTypeHandling: BadTypeHandling = BadTypeHandling(), _noLog: bool = False,
                        deserializeSerializable: bool = True, **deserializerKwargs) -> Any:
    """Deserialize a serialized field value. This is a recursive function able to traverse the type hint tree of a field,
    and deserialize it as needed.

    :param str fieldName: The name of the field. Used for more helpful error messages.
    :param fieldType: The type hint of the field in the class, as given by SerializableDataClass._typeOfFieldNamed
    :type fieldType: Union[type, _BaseGenericAlias, TypeVar, _MISSING_TYPE]
    :param PrimativeType serializedValue: The value to deserialize into fieldType
    :param VariableTrace c_variableTrace: A trace of the variables that the carica deserializer traversed to reach this variable
    :param BadTypeHandling c_badTypeHandling: How to handle receiving toml variables that do not match the type of the
                                                python variable. See class for default values and value descriptions.
    :param bool deserializeValues: Whether to automatically deserialize serialized Serializable fields (Default True)
    :raise TypeError: If a field has invalid type hints, or serializedValue does not match fieldType
    :return: serializedValue, deserialized into fieldType
    """
    def _log(*args, **kwargs):
        if not _noLog:
            log(*args, **kwargs)

    # take any type if type hinted as such
    if fieldType in (Any, object):
        return serializedValue

    # Handle Union type hints
    # Ignoring a warning here for missing attribute __origin__ - i just checked for it!
    elif hasattr(fieldType, "__origin__") and fieldType.__origin__ is Union: # type: ignore
        # Get the generic parameters
        genericArgs = cast(Tuple[FIELD_TYPE_TYPES_UNION, ...], typing.get_args(fieldType))
        # Make sure the Union was parameterised
        if len(genericArgs) == 0:
            raise TypeError(f"Field {fieldName} is a generic type but has not been parameterised")

        optional = any(genericType is type(None) for genericType in genericArgs)
        
        # Handle optional hints
        if optional and serializedValue is None:
            return None
        
        for genericType in genericArgs:
            if genericType is not type(None):
                try:
                    _deserializeField(fieldName, genericType, serializedValue, c_variableTrace=c_variableTrace,
                                        c_badTypeHandling=c_badTypeHandling, _noLog=True, **deserializerKwargs)
                except:
                    pass
                else:
                    return _deserializeField(fieldName, genericType, serializedValue, c_variableTrace=c_variableTrace,
                                                c_badTypeHandling=c_badTypeHandling, **deserializerKwargs)
       
        # We can't cast to a union, and potential casts to parameterized types are handled in the _deserializeField call
        # Therefore, skip CAST behaviour and jump straight to checking for KEEP
        fieldTypeError = f"Expected one of {'/'.join(str(t) for t in genericArgs)} for field {fieldName}, " \
                        + f"but received serialized type {type(serializedValue).__name__}"

        if c_badTypeHandling.behaviour == BadTypeBehaviour.KEEP:
            if c_badTypeHandling.logTypeKeeping:
                _log(f"[WARNING] Keeping mistyped field value: {fieldTypeError}")
            return serializedValue

        if c_badTypeHandling.rejectType == ErrorHandling.RAISE:
            raise TypeError(fieldTypeError)
        elif c_badTypeHandling.rejectType == ErrorHandling.LOG:
            _log(fieldTypeError)
        elif c_badTypeHandling.rejectType == ErrorHandling.IGNORE:
            pass

    # If type hinted with a normal type or protocol
    elif isinstance(fieldType, type):
        # Deserialize if needed
        if issubclass(fieldType, SerializableType):
            if deserializeSerializable:
                return fieldType.deserialize(serializedValue, **deserializerKwargs)
            else:
                return serializedValue

        # Otherwise do nothing
        elif issubclass(fieldType, primativeTypesTuple):
            if isinstance(serializedValue, fieldType):
                return _handleTypeCasts(serializedValue, fieldName, fieldType, c_badTypeHandling)
            raise TypeError(f"Expected type of {fieldType} for field {fieldName}, " \
                            + f"but received serialized type {type(serializedValue).__name__}")

    # Handle generics other than Union (e.g List)
    elif isinstance(fieldType, _BaseGenericAlias):
        # Get the generic (e.g List, Dict...)
        # ignoring a warning here on missing attribute - I can't import _BaseGenericAlias, so the type hinter doesn't know that __origin__ is guaranteed
        generic: _BaseGenericAlias = fieldType.__origin__ # type: ignore
        # If the type hint is like a dict
        if issubclass(generic, Dict):
            # Make sure the serialized value matches
            if not isinstance(serializedValue, Dict):
                raise TypeError(f"Received serialized value for field {fieldName} of type {type(serializedValue).__name__}" \
                                + ", but expected dict")
            # get the type parameters
            genericArgs = cast(Tuple[FIELD_TYPE_TYPES_UNION, ...], typing.get_args(fieldType))
            # make sure parameters were supplied
            if len(genericArgs) == 0:
                raise TypeError(f"Field {fieldName} is a generic type but has not been parameterised")
            # make sure we're only expecting str keys
            if isinstance(genericArgs[0], type) and not issubclass(genericArgs[0], str):
                raise exceptions.NonStringMappingKey(fieldType, path=c_variableTrace,
                                                        extra=f"Field {fieldName} is typed as a dict with non-str keys")

            # deserialize the dict
            newValue = {}
            for k, v in cast(Dict[str, Any], serializedValue).items():
                newValue[k] = _deserializeField(fieldName, genericArgs[1], v,
                                                c_variableTrace=c_variableTrace + [k], c_badTypeHandling=c_badTypeHandling,
                                                **deserializerKwargs)
            return newValue

        # If the type hint is like a collection
        elif issubclass(generic, (List, Set, Tuple)):
            # Make sure the serialized type matches
            if not isinstance(serializedValue, (List, Set, Tuple)):
                raise TypeError(f"Expected type of {fieldType} for field {fieldName}, " \
                            + f"but received serialized type {type(serializedValue).__name__}")

            # get the type parameters
            genericArgs = cast(Tuple[FIELD_TYPE_TYPES_UNION, ...], typing.get_args(fieldType))
            # make sure parameters were supplied
            if len(genericArgs) == 0:
                raise TypeError(f"Field {fieldName} is a generic type but has not been parameterised")
            # make sure only one type was supplied
            if len(genericArgs) > 1 and genericArgs[1] is not Ellipsis:
                raise TypeError(f"Field {fieldName} is typed as a tuple with multiple type slots. " \
                                + "Tuples must be parameterised with a single type, or a type followed by ...")
            
            # deserialize each element in the collection in turn
            builder = (_deserializeField(fieldName, genericArgs[0], v, c_variableTrace=c_variableTrace + [i],
                                        c_badTypeHandling=c_badTypeHandling, **deserializerKwargs)
                        for i, v in enumerate(serializedValue)) # type: ignore
            hintedTypes = {List: list, Set: set, Tuple: tuple}
            if generic in hintedTypes:
                return hintedTypes[generic](builder)
            return generic(builder)
            
    # Failed to deserialize
    raise TypeError(f"Field {fieldName} is typed as a non-serializable type: {fieldType}")



@dataclass(init=True, repr=True, eq=True)
class SerializableDataClass(SerializesToDict):
    """An dataclass with added serialize/deserialize methods.
    Values stored in the fields of the dataclass are not type checked, but must be primatives/serializable for the serialize
    method to return valid results.

    Subclasses of SerializableDataClass *must* be decorated with `@dataclasses.dataclass` to function properly.
    """
    @classmethod
    def _getFields(cls) -> Dict[str, Field]:
        """Get the `@dataclass`-generated mapping of field names to `Field`s.
        Separating this method out to isolate mypy warnings.

        :return: A dictionary mapping field names to `dataclasses.Field`
        :rtype: Dict[str, Field]
        """
        return {field.name: field for field in fields(cls)}


    @classmethod
    def _fieldNames(cls) -> List[str]:
        """Get a list of the field names defined in this class.

        :return: A list of the field names defined in the class.
        :rtype: List[str]
        """
        return list(cls._getFields().keys())


    def _fieldItems(self):
        """Get a `dict.items()`-style mapping of field names to field values.

        :return: a dictionary mapping field names to current values
        :rtype: Dict[str, Any]
        """
        return {k: getattr(self, k) for k in self._fieldNames()}.items() #self.__dataclass_fields__


    @classmethod
    def _typeOfFieldNamed(cls, fieldName: str) -> Union[type, _BaseGenericAlias, TypeVar, _MISSING_TYPE]:
        """Get the type annotation for the field with the given name. This does not consider type overriding.
        Be aware that type-hinting will cause this function to return a 'typing' type.

        :return: The type of the field called `fieldName`, or `dataclasses._MISSING_TYPE` if no type was given for the field
        :rtype: type
        """
        return cls._getFields()[fieldName].type


    @classmethod
    def _overriddenTypeOfFieldNamed(cls, fieldName: str) -> Union[type, _BaseGenericAlias, TypeVar, _MISSING_TYPE]:
        """Get the type annotation for the field with the given name, with awareness for overridden types.
        Be aware that type-hinting will cause this function to return a 'typing' type.

        :return: The (possibly overridden) type of the field called `fieldName`, or `dataclasses._MISSING_TYPE` if no type was given for the field
        :rtype: type
        """
        f = cls._getFields()[fieldName]
        if cls._fieldTypeIsOverridden(fieldName):
            if isinstance(f.default_factory, _CallableDeserializedTypeOverrideProxy):
                return cast(_CallableDeserializedTypeOverrideProxy, f.default_factory)._self__carica_uninitialized_type__
            return cast(_DeserializedTypeOverrideProxy, f.default)._self__carica_uninitialized_type__
        return f.type


    @classmethod
    def _fieldTypeIsOverridden(cls, fieldName: str) -> bool:
        """Decide whether the deserialized type for a field has been overridden.

        :return: True if the field called `fieldName` was marked with `TypeOverride`, False otherwise
        :rtype: type
        """
        f = cls._getFields()[fieldName]
        return isinstance(f.default, _DeserializedTypeOverrideProxy) or \
                isinstance(f.default_factory, _CallableDeserializedTypeOverrideProxy)


    @classmethod
    def _hasISerializableOrGenericField(cls) -> bool:
        """Decide whether this class has any fields which are Serializable or generic types.
        If any fields are themselves Serializable types, then we know that we need to traverse the field types tree when
        deserializing to receive proper types. If any fields are generic, then we do not immediately know if any Serializable
        generics exist without traversing the field types tree. If we're doing that, then we might as well deserialize while
        we do it!

        :return: False if we are immediately sure no fields contain Serializable types, True otherwise
        :rtype: bool
        """
        for fieldName in cls._fieldNames():
            fieldType = cls._typeOfFieldNamed(fieldName)
            if not isinstance(fieldType, _MISSING_TYPE):
                if isinstance(fieldType, type):
                    if issubclass(fieldType, SerializableType):
                        return True

                elif isinstance(fieldType, _BaseGenericAlias):
                    return True

                elif isinstance(fieldType, TypeVar):
                    if isinstance(fieldType, SerializableType):
                        return True
        
        return False

    
    def serialize(self, deepTypeChecking: bool = False, **kwargs) -> Dict[str, PrimativeType]:
        """Serialize this object into a dictionary, to be recreated completely.

        :param bool deepTypeChecking: Whether to ensure serializability of member objects recursively (Default False)
        :return: A dictionary mapping field names to serialized values
        :rtype: Dict[str, PrimativeType]
        """
        data: Dict[str, PrimativeType] = {}

        for k in self._getFields():
            v = getattr(self, k)
            if isinstance(v, SerializableType):
                data[k] = v.serialize(**kwargs)
            elif not objectIsShallowSerializable(v):
                raise exceptions.NonSerializableObject(v, extra=f"Field '{k}' is of non-serializable type: {type(v).__name__}")
            elif deepTypeChecking and not objectIsDeepSerializable(v):
                raise exceptions.NonSerializableObject(v, extra=f"Field '{k}' has a non-serializable member object")
            else:
                data[k] = v

        return data


    @classmethod
    def deserialize(cls, data: Mapping[str, PrimativeType], deserializeValues: bool = True, c_variableTrace: VariableTrace = [], **kwargs):
        """Recreate a serialized SerializableDataClass object. If `deserializeValues` is `True`,
        values fields which are serializable types will be automatically deserialized.

        :param Dict[str, PrimativeType] data: A dictionary mapping field names to values.
        :param bool deserializeValues: Whether to automatically deserialize serialized Serializable fields (Default True)
        :return: A new object as specified by data
        :rtype: SerializableDataClass
        """
        if not isinstance(data, dict):
            raise TypeError(f"Invalid serialized {cls.__name__}. Expected Dict[str, PrimativeType], received {type(data).__name__}")

        for k, v in data.items():
            if not isinstance(k, str):
                raise exceptions.NonStringMappingKey(k, path=c_variableTrace)
            data[k] = _deserializeField(k, cls._overriddenTypeOfFieldNamed(k), v, c_variableTrace=c_variableTrace + [k], 
                                        deserializeSerializable=deserializeValues, **kwargs)

        constructorArgs = inspect.signature(cls.__init__).parameters
        classKwargs = {k: v for k, v in kwargs.items() if k in constructorArgs}

        return cls(**data, **classKwargs)



from typing import Any, Optional, List, Union


class KeyTracedException(Exception):
    """Base exception type for exceptions which occur when scanning potentially recursive types.
    
    :ivar o: The erroneous object
    :vartype o: Any
    :ivar depth: If the error occured when studying an object inside of an iterable or mapping, this represents the number
                of objects deep where the erroneous object was located. E.g in [[x]], x is 2 layers deep
    :vartype depth: Optional[int]
    :ivar path:  If the error occured when studying an object inside of an iterable or mapping, this path should trace
                the keys/indices accessed to reach the object
    :vartype path: Optional[List[Union[str, int]]]
    :ivar extra: Any extra details which may be useful for debugging or locating the erroneous object
    :vartype extra: str
    """
    
    def __init__(self, o: Any, depth: Optional[int] = None, path: Optional[List[Union[str, int]]] = None,
                        extra: Optional[str] = None) -> None:
        """
        :param o: The erroneous object
        :type o: Any
        :param depth: If the error occured when studying an object inside of an iterable or mapping, this represents the number
                of objects deep where the erroneous object was located. E.g in [[x]], x is 2 layers deep
        :type depth: Optional[int]
        :param path:  If the error occured when studying an object inside of an iterable or mapping, this path should trace
                    the keys/indices accessed to reach the object (Default None)
        :type path: Optional[List[Union[str, int]]]
        :param extra: Any extra details which may be useful for debugging or locating the erroneous object (Default None)
        :type extra: str
        """
        self.o = o
        if depth is None and path is not None:
            depth = len(path)
        self.depth = depth
        self.path = path
        self.extra = extra
        super().__init__()


    def formatPathInfo(self) -> str:
        """Format the exception's depth, path and extra info into a single string, if they exist.

        :return: A nicely formatted string with the depth, path and extra info of the exception, if they exist
        :rtype: str
        """
        info = ""
        if self.depth:
            info += f"at depth {self.depth}"
        if self.path:
            info += f"{' ' if self.depth else ''}({'->'.join(str(i) for i in self.path)})"
        if self.extra:
            info += f"{' ' if self.depth or self.path else ''}({self.extra})"
        
        return info


    def __str__(self) -> str:
        return f"object '{repr(self.o)}' {self.formatPathInfo()}"


class NonSerializableObject(KeyTracedException):
    """Exception to be raised when attempting to serialize an object of a non-serializable type.
    """
    pass


class NonStringMappingKey(KeyTracedException):
    """Exception to be raised when attempting to serialize a mapping (dict) which has a key that is not a str.
    """
    pass


class MultiTypeList(KeyTracedException):
    """Exception to be raised when attempting to serialize a list containing both tables (dicts) and non-table types.
    """
    def __str__(self) -> str:
        return f"List contains both dicts and non-dict types {self.formatPathInfo()}"



from typing import Any, Dict, List, Union, cast
from tomlkit.container import Container as TKContainer
from tomlkit.items import AoT

INCOMPATIBLE_TOML_TYPES = (TKContainer, AoT)

def convertIncompatibleTomlTypes(doc: Union[TKContainer, AoT]) -> Union[List[Any], Dict[str, Any]]:
    """Recursively converts tomlkit tables/lists of tables into dictionaries/lists of dictionaries

    :param doc: The item to convert
    :type doc: Union[TKContainer, AoT]
    :return: doc recursively converted to python types instead of TOMLKit Containers/AoTs
    """
    if isinstance(doc, TKContainer):
        # Not sure why mypy doesn't pick up on Container inheriting from dict
        return {k: convertIncompatibleTomlTypes(v) for k, v in  cast(dict, doc).items()}
    elif isinstance(doc, AoT):
        return [convertIncompatibleTomlTypes(v) for v in doc]
    else:
        return doc



from typing import Any, Iterable, List, Mapping, TypeVar, Union, cast, Callable
# ignoring a warning here because private type _BaseGenericAlias can't be imported right now.
# it is a necessary import to unify over user-defined and special generics.
from typing import _BaseGenericAlias # type: ignore
from carica.interface import serializableTypesTuple, primativeTypesTuple
from carica import exceptions
from wrapt import ObjectProxy, CallableObjectProxy

TypeHint = Union[type, _BaseGenericAlias]

class _DeserializedTypeOverrideProxy(ObjectProxy):
    def __init__(self, wrapped, typeOverride: TypeHint):
        super().__init__(wrapped)
        self._self__carica_uninitialized_type__ = typeOverride

    def __repr__(self):
        inner = repr(self.__wrapped__)
        if inner.endswith(">"):
            return f"{inner[:-1]}[Carica-TypeOverride-Proxy]>"
        return f"<{inner}[CaricaTypeOverrideProxy]>"

class _CallableDeserializedTypeOverrideProxy(_DeserializedTypeOverrideProxy, CallableObjectProxy): pass

T = TypeVar("T")
def TypeOverride(override: TypeHint, defaultValue: T) -> T:
    if isinstance(defaultValue, Callable):
        return cast(T, _CallableDeserializedTypeOverrideProxy(defaultValue, override))
    return cast(T, _DeserializedTypeOverrideProxy(defaultValue, override))

def objectIsObjectIterable(o: Any) -> bool:
    """Decide whether o is an iterable of objects.
    like typing.Iterable, but more restricted for serializabiliy checking.

    :return: True if o is of a type considered to be an iterable of objects, but not a string
    :rtype: False
    """
    return isinstance(o, Iterable) and not isinstance(o, str)


def objectIsShallowPrimative(o: Any) -> bool:
    """Decide if an object is a primative type, excluding any objects potentially contained within it.
    I.e:
    ```
    >>> objectIsShallowPrimative(1)
    True
    >>> class MyNonPrimativeClass:
    ...    pass
    ...
    >>> objectIsShallowPrimative(MyNonPrimativeClass())
    False
    >>> objectIsShallowPrimative([MyNonPrimativeClass()])
    True
    ```

    :param Any o: The object whose type to check
    :return: True if o itself is a primative type, False otherwise
    :rtype: bool
    """
    return isinstance(o, primativeTypesTuple)


def objectIsDeepPrimative(o: Any) -> bool:
    """Decide if a class is a primative type, including any objects potentially contained within it.
    I.e:
    ```
    >>> objectIsDeepPrimative(1)
    True
    >>> objectIsDeepPrimative([1])
    True
    >>> class MyNonPrimativeClass:
    ...    pass
    ...
    >>> objectIsDeepPrimative(MyNonPrimativeClass())
    False
    >>> objectIsDeepPrimative([MyNonPrimativeClass()])
    False
    ```

    This process is highly inefficient, as it recurses through all member objects in `o`.
    For a more efficient (but less confident) check, use `objectIsShallowPrimative` to check if just `o` itself is primative,
    excluding any member objects.

    :param Any o: The object whose type to check
    :return: True if o itself is a primative type, False otherwise
    :rtype: bool
    """
    if objectIsObjectIterable(o):
        return all(objectIsDeepPrimative(i) for i in o)
    elif isinstance(o, Mapping):
        return all(isinstance(k, str) and objectIsDeepPrimative(v) for k, v in o.items())
    else:
        return objectIsShallowPrimative(o)


def objectIsShallowSerializable(o: Any) -> bool:
    """Decide if an object is a primative or serializable type, excluding any objects potentially contained within it.
    I.e:
    ```
    >>> objectIsShallowSerializable(1)
    True
    >>> class MySerializableClass:
    ...    def serialize(self, **kwargs):
    ...        return {}
    ...    def deserialize(self, data **kwargs):
    ...        return MySerializableClass()
    ...
    >>> objectIsShallowSerializable(MySerializableClass())
    True
    >>> class MyNonPrimativeClass:
    ...    pass
    ...
    >>> objectIsShallowSerializable(MyNonPrimativeClass())
    False
    >>> objectIsShallowSerializable([MyNonPrimativeClass()])
    True
    ```

    :param Any o: The object whose type to check
    :return: True if o itself is a primative or serializable type, False otherwise
    :rtype: bool
    """
    return isinstance(o, serializableTypesTuple)


def objectIsDeepSerializable(o: Any) -> bool:
    """Decide if a class is a primative or serializable type, including any objects potentially contained within it.
    I.e:
    ```
    >>> objectIsDeepSerializable(1)
    True
    >>> class MySerializableClass:
    ...    def serialize(self, **kwargs):
    ...        return {}
    ...    def deserialize(self, data **kwargs):
    ...        return MySerializableClass()
    ...
    >>> objectIsDeepSerializable(MySerializableClass())
    True
    >>> objectIsDeepSerializable([MySerializableClass()])
    True
    >>> class MyNonPrimativeClass:
    ...    pass
    ...
    >>> objectIsDeepSerializable(MyNonPrimativeClass())
    False
    >>> objectIsDeepSerializable([MyNonPrimativeClass()])
    False
    ```

    This process is highly inefficient, as it recurses through all member objects in `o`.
    For a more efficient (but less confident) check, use `objectIsShallowPrimative` to check if just `o` itself is primative,
    excluding any member objects.

    :param Any o: The object whose type to check
    :return: True if o itself is a primative type, False otherwise
    :rtype: bool
    """
    if objectIsObjectIterable(o):
        return all(objectIsDeepSerializable(i) for i in o)
    elif isinstance(o, Mapping):
        return all(isinstance(k, str) and objectIsDeepSerializable(v) for k, v in o.items())
    else:
        return objectIsShallowSerializable(o)


def raiseForShallowNonSerializable(o: Any) -> None:
    """The same as objectIsShallowSerializable, but with different return semantics:
    If `o` is shallow serializable, None is returned.
    If `o` is shallow non-serializable an exception is raised with details of the reasoning.

    :param Any o: The object whose type to check
    :raise exceptions.NonSerializableObject: If o is itself a non-serializable type
    :raise exceptions.NonStringMappingKey: If o is a mapping containing a non-str key
    """
    if isinstance(o, Mapping):
        for k in o:
            if not isinstance(k, str):
                raise exceptions.NonStringMappingKey(k)
    elif not objectIsShallowSerializable(o):
        raise exceptions.NonSerializableObject(o)


def _recurseRaiseForDeepNonSerializable(o: Any, path: List[Union[str, int]]) -> None:
    """Internal recursive method for use exclusively by raiseForDeepNonSerializable.
    """
    if objectIsObjectIterable(o):
        for i in o:
            _recurseRaiseForDeepNonSerializable(i, path + [i])
    elif isinstance(o, Mapping):
        for k, v in o.items():
            if not isinstance(k, str):
                raise exceptions.NonStringMappingKey(k, len(path), path)
            _recurseRaiseForDeepNonSerializable(v, path + [k])
    else:
        if not objectIsShallowSerializable(o):
            raise exceptions.NonSerializableObject(o, len(path), path)


def raiseForDeepNonSerializable(o: Any) -> None:
    """The same as objectIsDeepSerializable, but with different return semantics:
    If `o` is completely serializable, None is returned.
    If `o` is non-serializable or contains a non-serializable member object, an exception is raised with details of
    the reasoning.

    This process is highly inefficient, as it recurses through all member objects in `o`.
    For a more efficient (but less confident) check, use `raiseForShallowNonSerializable` to check if just `o` itself is
    serializable, excluding any member objects.

    :param Any o: The object whose type to check
    :raise exceptions.NonSerializableObject: If a non-serializable object is encountered at any level
    :raise exceptions.NonStringMappingKey: If a non-str mapping key is encountered at any level
    """
    _recurseRaiseForDeepNonSerializable(o, [])



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Generic, Iterable, Mapping, Type, TypeVar, Union, Protocol, runtime_checkable, Optional
from datetime import datetime

# All types acceptable as toml data. Tomlkit handles serializing of datetime objects automatically.
# Iterable includes set, list and tuple. It also includes dict and str!
# this type is Optional, because None is allowed in toml, but there is no NoneType exposed in python for use in the Union.
PrimativeType = Optional[Union[int, float, str, bool, datetime, Iterable["PrimativeType"], Mapping[str, "PrimativeType"]]]
# PrimativeType as a shallow set
primativeTypes = {int, float, str, bool, Iterable, Mapping, datetime, type(None)}
# PrimativeTypes as a shallow tuple
primativeTypesTuple = tuple(primativeTypes)

TClass = TypeVar("TClass")

@runtime_checkable
class SerializableType(Protocol):
    """A type protocol representing any object with `serialize` and `deserialize` methods defined.
    Since this is a runtime checkable protocol class, isinstance() and issubclass() work for duck typing. That is,
    they will return true for classes with `serialize` and `deserialize` defined, even if the class does not inherit
    from `SerializableType`. For inheritence purposes, you may instead wish to use the `ISerializable` interface.
    """
    
    def serialize(self, **kwargs) -> PrimativeType:
        """Serialize this object into primative types (likely a dictionary, e.g JSON), to be recreated completely.

        :return: A primative (likely a dictionary) containing all information needed to recreate this object
        :rtype: PrimativeType
        """
        ...


    @classmethod
    def deserialize(cls: Type[TClass], data: PrimativeType, **kwargs) -> TClass:
        """Recreate a serialized Serializable-like object

        :param PrimativeType data: A primative (likely a dictionary) containing all information needed to recreate the serialized object
        :return: A new object as specified by data
        :rtype: Serializable-like
        """
        ...
        

class ISerializable(ABC):
    """An object which can be represented entirely by a dictionary of primitives, created with the serialize method.
    This object can then be recreated perfectly using the deserialize method.
    """
    
    @abstractmethod
    def serialize(self, **kwargs) -> PrimativeType:
        """Serialize this object into primative types (likely a dictionary, e.g JSON), to be recreated completely.

        :return: A primative (likely a dictionary) containing all information needed to recreate this object
        :rtype: PrimativeType
        """
        raise NotImplementedError()


    @classmethod
    @abstractmethod
    def deserialize(cls: Type[TClass], data: PrimativeType, **kwargs) -> TClass:
        """Recreate a serialized ISerializable object

        :param PrimativeType data: A primative (likely a dictionary) containing all information needed to recreate the serialized object
        :return: A new object as specified by data
        :rtype: ISerializable
        """
        raise NotImplementedError()
    

# All types which are themselves primative, or can be serialized into primative types, as a shallow set
serializableTypes = primativeTypes.copy()
serializableTypes.add(SerializableType)
serializableTypesTuple = tuple(serializableTypes)

TClass = TypeVar("TClass")
TSerialized = TypeVar("TSerialized", bound=PrimativeType)

class SerializesToType(SerializableType, Generic[TSerialized]):
    def serialize(self, **kwargs) -> TSerialized: ...

    @classmethod
    def deserialize(cls: Type[TClass], data: TSerialized, **kwargs) -> TClass: ...

SerializesToDict = SerializesToType[Mapping[str, PrimativeType]]



from .carica import makeDefaultCfg, loadCfg, BadTypeBehaviour, BadTypeHandling, ErrorHandling
from .typeChecking import TypeOverride
from .models import *
from .interface import *
from . import exceptions

__version__ = '1.3.2'



from typing import Type, TypeVar
from carica.interface import ISerializable, PrimativeType
from pathlib import Path, WindowsPath, PosixPath
import os

TSelf = TypeVar("TSelf", bound="SerializablePath")

class SerializablePath(ISerializable, Path):
    """A serializable path intended to be treated as a string.
    This class serializes into a usable string.

    This class directly copies the instancing logic of pathlib.Path in order to achieve subclassing.
    For more details, see the implementation of pathlib.Path, and this stackoverflow thread (which has not been followed):
    https://stackoverflow.com/questions/29850801/subclass-pathlib-path-fails
    """

    def __new__(cls, *args, **kwargs) -> "SerializablePath":
        if cls is SerializablePath:
            cls = SerializableWindowsPath if os.name == 'nt' else SerializablePosixPath
        # ignoring a warning here on missing method _from_parts
        # the OS-specific serializable path subclasses still transitively extend Path, so _from_parts is guaranteed
        self = cls._from_parts(args, init=False) # type: ignore[reportGeneralTypeIssues]
        if not self._flavour.is_supported:
            raise NotImplementedError("cannot instantiate %r on your system"
                                      % (cls.__name__,))
        self._init()
        return self


    def serialize(self, **kwargs) -> str:
        """Return this path as a string.

        :return: The path as a string
        :rtype: str
        """
        return str(self)


    @classmethod
    def deserialize(cls: Type[TSelf], data: PrimativeType, **kwargs) -> TSelf:
        """Form a SerializablePath from a string. Separation of the path into parts is handled by pathlib.Path logic.

        :param str data: The path to deserialize, as an os.sep-separated string of path parts.
        :return: A new SerializablePath representing data
        :rtype: SerializablePath
        """
        if not isinstance(data, str):
            raise TypeError(f"Invalid type for parameter data. Expected str, received {type(data).__name__}")
        return cls(data)


    def __add__(self: TSelf, other: object) -> TSelf:
        """Concatenate two paths into a new SerializablePath. Neither this nor other are modified.

        :param other: The path to appear on the right-hand side of the new path
        :type other: Union[str, Path, SerializablePath]
        :return: A new SerializablePath containing this path followed by other, separated with os.sep as necessary
        """
        if isinstance(other, str):
            return type(self)(*(self.parts + (other,)))
        elif isinstance(other, Path):
            return type(self)(*(self.parts + other.parts))
        else:
            raise TypeError(f"Incompatible types: {type(self).__name__} and {type(other).__name__}")


class SerializableWindowsPath(SerializablePath, WindowsPath):
    pass

class SerializablePosixPath(SerializablePath, PosixPath):
    pass



from typing import Dict, TypeVar, Type
from carica.interface.Serializable import ISerializable, PrimativeType
from datetime import timedelta

TSelf = TypeVar("TSelf", bound="SerializableTimedelta")

class SerializableTimedelta(ISerializable, timedelta):
    """A serializable version of `datetime.timedelta`. For `datetime.datetime`, no extra handling is needed,
    as datetime is considered a primitive type by TOML.
    """
    @classmethod
    def fromTimedelta(cls, td: timedelta):
        """Create a new `SerializableTimedelta` from an existing `datetime.timedelta`.

        :param timedelta td: The timedelta to duplicate
        """
        return cls(days=td.days, seconds=td.seconds, microseconds=td.microseconds)


    def serialize(self, **kwargs) -> Dict[str, int]:
        """Serialize this timedelta into the simplest representation possible - using the largest amounts of the largest
        units that will fit into this timedelta.

        :return: A dictionary containing all units, each possibly zero, to represent this timedelta.
        :rtype: Dict[str, int]
        """
        data = {}

        data["weeks"] = 0 if self.days < 7 else self.days // 7
        data["days"] = self.days - data["weeks"] * 7

        data["hours"] = 0 if self.seconds < 3600 else self.seconds // 3600
        seconds = self.seconds - data["hours"] * 3600
        data["minutes"] = 0 if seconds < 60 else seconds // 60
        data["seconds"] = seconds - data["minutes"] * 60
        
        data["milliseconds"] = 0 if self.microseconds < 1000 else self.microseconds // 1000
        data["microseconds"] = self.microseconds - data["milliseconds"] * 1000
            
        return data


    @classmethod
    def deserialize(cls: Type[TSelf], data: PrimativeType, **kwargs) -> TSelf:
        if not isinstance(data, dict):
            raise TypeError(f"Invalid serialized {cls.__name__}: {data}")
        return cls(**data)



from .Serializable import ISerializable, SerializableType, \
                            PrimativeType, primativeTypes, primativeTypesTuple, serializableTypes, \
                            serializableTypesTuple, primativeTypesTuple, SerializesToDict, SerializesToType


from enum import Enum
from types import ModuleType
import tomlkit
from tomlkit.container import Container as TKContainer
from tomlkit import items as TKItems
import os
from typing import Any, Dict, List, Protocol, Union, Iterable, Mapping, cast, runtime_checkable
import tokenize

from tomlkit.toml_document import TOMLDocument
from carica.interface import SerializableType, PrimativeType
from carica.typeChecking import objectIsShallowPrimative, _DeserializedTypeOverrideProxy, TypeHint
from carica import exceptions
from carica.util import INCOMPATIBLE_TOML_TYPES, convertIncompatibleTomlTypes
from dataclasses import dataclass
import traceback

CFG_FILE_EXT = ".toml"
DISALLOWED_TOKEN_TYPES = {tokenize.INDENT}
DISALLOWED_TOKEN_TYPE_VALUES = {tokenize.NAME: {"from", "import"}}
LINE_DELIMITER_TOKEN_TYPES = {tokenize.NL, tokenize.NEWLINE, tokenize.ENDMARKER}
IGNORED_TOKEN_TYPES = {tokenize.DEDENT}

VariableTrace = List[Union[int, str]]


def log(msg):
    print(msg)


def formatException(e, includeTrace):
    if includeTrace:
        return "".join(traceback.format_exception(type(e), e, e.__traceback__))
    return str(e)


def tokenDisallowed(token: tokenize.TokenInfo) -> bool:
    """Decide whether or not a token can be present in avariable assignment line.
    The return value is the inverse of what you would expect: True if the token can NOT be part of a variable assignment line.

    :param tokenize.TokenInfo token: The token to check
    :return: True if the token is an ILLEGAL part of a variable assignment, False if the token is LEGAL.
    :rtype: bool
    """
    return token.type in DISALLOWED_TOKEN_TYPES or \
        (token.type in DISALLOWED_TOKEN_TYPE_VALUES and token.string in DISALLOWED_TOKEN_TYPE_VALUES[token.type])


def lineStartsWithVariableIdentifier(line: List[tokenize.TokenInfo]) -> bool:
    """Decide if a series of tokens starts with a variable identifier.
    This is a large estimation.
    """
    return len(line) > 1 \
        and line[0].type == tokenize.NAME \
        and line[1].type == tokenize.OP and line[1].string == "="


@dataclass
class ConfigVariable:
    name: str
    value: Any
    inlineComments: List[str]
    preComments: List[str]
    loadedType: TypeHint


    def hasInline(self):
        return self.inlineComments != []


    def hasPre(self):
        return self.preComments != []


@runtime_checkable
class _TKItemWithValue(Protocol):
    def value(self): ...


def _partialModuleVariables(module: ModuleType) -> Dict[str, ConfigVariable]:
    """Estimate The set of variables defined in the given module.
    Attempts to ignore other name types, such as functions, classes and imports.
    Names cannot be indented, and multiple assignment is not supported - variable declarations must be on their own lines.

    The resulting dictionary maps variable names to a data class containing the variable name, value, inline comment(s),
    and preceeding comments. Inline comments can be multiple if the variable was assigned to a second time with a new inline
    comment. A comment 'preceeds' a variable if it is present either on the line before the variable, or on the line before
    a preceeding comment of the variable.

    This function simply iterates over the module's tokens, and does not build an AST.
    This means that certain name definition structures will result in false positives/negatives.
    This behaviour has not been extensively tested, but once such false positive has been identified:
    When invoking a callable (such as a class or function) with a keyword argument on a new, unindented line, the argument
    name will be falsely identified as a variable name. E.g:
    ```
    my_variable = dict(key1=value1,
    key2=value2)
    ```
    produces `my_variable` and `key2` as variable names.

    :param ModuleType module: The module from which to estimate variable names
    :return: An estimated dict of variables defined in module. See above for details and limitations.
    :rtype: Dict[str, ConfigVariable]
    """
    # TODO: Ignoring a warning here because virtual modules are not supported yet
    with tokenize.open(module.__file__) as f: # type: ignore[reportGeneralTypeIssues]
        tokens = tokenize.generate_tokens(f.readline)
        # All detected variables
        moduleVariables: Dict[str, ConfigVariable] = {}
        # All tokens on the current line
        currentLine: List[tokenize.TokenInfo] = []
        # Track consecutive lines of comments, in case they preceed a variable
        commentsQueue: List[str] = []

        ignoreLine = False
        
        # Iterate over all tokens in the file. tokenizer doesn't split by new line for us, so we'll have to do it manually.
        for token in tokens:
            # If the current line is to be ignored
            if ignoreLine:
                # Continue iterating until a new line character is found, then stop and go back to normal parsing
                if token.type == tokenize.NEWLINE:
                    ignoreLine = False
                continue

            # Filter out lines that start with a token that eliminates the possiblity of this line defining a variable
            if currentLine == [] and tokenDisallowed(token):
                # Set line ignore flag
                if token.type != tokenize.NEWLINE:
                    ignoreLine = True
                # Clear the preceeding comments queue
                if commentsQueue:
                    commentsQueue.clear()
                # Proceed to the next line of code
                continue
            
            # End of line reached?
            elif token.type in LINE_DELIMITER_TOKEN_TYPES:
                # Were there any tokens on the line?
                if currentLine:
                    # This check needs to be performed before clearing the preceeding comments queue,
                    # As it checks if the new line breaks a series of comments. I might as well put the
                    # variable registering logic in here as well, because it also can only happen if the
                    # line did not start with a comment, and it also calls for clearing the preeding comments queue.
                    if currentLine[0].type != tokenize.COMMENT:
                        # Was the first token on the line a variable identifier? *This is the estimation bit*
                        if lineStartsWithVariableIdentifier(currentLine):
                            # Store a record of the variable
                            variableName = currentLine[0].string
                            if variableName not in moduleVariables:
                                # Plaster solution to ignore false positives
                                try:
                                    variableValue = getattr(module, variableName)
                                except AttributeError:
                                    # This variable name must be a false positive.
                                    # Clear the preceeding comments queue, and move onto the next line
                                    if commentsQueue:
                                        commentsQueue.clear()
                                    currentLine = []
                                    continue
                                
                                # Variable value was fetched successfully, record it
                                loadedType = variableValue._self__carica_uninitialized_type__ if isinstance(variableValue, _DeserializedTypeOverrideProxy) else type(variableValue)
                                moduleVariables[variableName] = ConfigVariable(variableName, variableValue, [], [], loadedType)

                            # Record any preceeding comments
                            if commentsQueue:
                                moduleVariables[variableName].preComments += commentsQueue

                        # Clear the preceeding comments queue
                        if commentsQueue:
                            commentsQueue.clear()

                    # Move onto the next line
                    currentLine = []
                
                # No tokens on the line, breaking any series of comments. Clear the preceeding comments queue
                elif commentsQueue:
                    commentsQueue.clear()

            # Is this a comment?
            elif token.type == tokenize.COMMENT:
                # Is it an inline comment?
                if currentLine != []:
                    # Does this line declare a variable?
                    if lineStartsWithVariableIdentifier(currentLine):
                        variableName = currentLine[0].string

                        # Create a record for the variable if none exists
                        if variableName not in moduleVariables:
                            variableValue = getattr(module, variableName)
                            loadedType = variableValue._self__carica_uninitialized_type__ if isinstance(variableValue, _DeserializedTypeOverrideProxy) else type(variableValue)
                            moduleVariables[variableName] = ConfigVariable(variableName, variableValue, [], [], loadedType)

                        # Record the found inline comment
                        moduleVariables[variableName].inlineComments.append(token.string.lstrip("# "))
                
                # Is it a comment line?
                else:
                    # Add the comment text to the preceeding comments queue
                    commentsQueue.append(token.string.lstrip("# "))
                    currentLine.append(token)

            # No processing of this token to be performed, record it and move onto the next one in the line
            elif token.type not in IGNORED_TOKEN_TYPES:
                currentLine.append(token)
        
    return moduleVariables


def _serialize(o: Any, path: VariableTrace, depthLimit=20, serializerKwargs={}) -> PrimativeType:
    """Internal recursive method to serialize any serializable object, or throw exceptions with useful key trace info
    """
    # Check recursion depth
    if len(path) > depthLimit:
        raise RecursionError()

    # o is directly serializable, serialize it and return that
    if isinstance(o, SerializableType):
        return _serialize(o.serialize(**serializerKwargs), path + ["[serialize]"], depthLimit=depthLimit,
                            serializerKwargs=serializerKwargs)

    # o is a mapping, serialize each value
    elif isinstance(o, Mapping):
        serializedDict: Dict[str, PrimativeType] = {}
        for k, v in o.items():
            # ensure keys are str
            if not isinstance(k, str):
                raise exceptions.NonStringMappingKey(k, len(path)-1, path)

            serializedDict[k] = _serialize(v, path + [k], depthLimit=depthLimit, serializerKwargs=serializerKwargs)
        return serializedDict

    # o is an iterable, serialize each element
    if isinstance(o, Iterable) and not isinstance(o, str):
        serializedList: List[PrimativeType] = []
        for i in o:
            serializedList.append(_serialize(i, path + [i], depthLimit=depthLimit, serializerKwargs=serializerKwargs))
        
        # ensure serialized lists contain either tables or non-dict primatives, but not both
        if len(serializedList) > 1:
            if (isinstance(serializedList[0], Mapping) and any(not isinstance(i, Mapping) for i in serializedList[1:])) or \
                    (not isinstance(serializedList[0], Mapping) and any(isinstance(i, Mapping) for i in serializedList[1:])):
                raise exceptions.MultiTypeList(o, len(path)-1, path)

        return serializedList

    # o is just a normal primative, return it as is
    elif objectIsShallowPrimative(o):
        return o

    # unable to match serializing method, exception time
    raise exceptions.NonSerializableObject(o, len(path)-1, path)


def makeDefaultCfg(cfgModule: ModuleType, fileName: str = "defaultCfg" + CFG_FILE_EXT, retainComments: bool = True,
                    **serializerKwargs) -> str:
    """Create a config file containing all configurable variables with their default values.
    The name of the generated file may optionally be specified.

    fileName may also be a path, either relative or absolute. If missing directories are specified
    in fileName, they will be created.

    If fileName already exists, then the generated file will be renamed with an incrementing number extension.

    This method has the option to retain variable 'docstrings' - comments which Carica deems to belong to a variable.
    A comment may belong to a variable if it is one of:
        a) On the same line as the variable declaration (an 'inline comment')
        b) A line comment on the line immediately preceeding the variable declaration (a 'preceeding comment')
        c) A line comment on the line immediately preceeding an existing preceeding comment, allowing for multi-line comments

    Retaining of preceeding comments is currently disabled. This is due to the TOML spec's standards for variable ordering
    within a document being stricter than python's, leading to list and dict preceeding comments appearing in odd places.
    This is currently in the process of being worked around.

    :param ModuleType cfgModule: Module to convert to toml
    :param str fileName: Path to the file to generate (Default "defaultCfg.toml")
    :param bool retainComments: Whether or not to write variable docstrings to the config (Default True)
    :return: path to the generated config file
    :rtype: str
    :raise ValueError: If fileName does not end in CFG_FILE_EXT
    """
    # Ensure fileName is toml
    if not fileName.endswith(CFG_FILE_EXT):
        raise ValueError("file name must end with " + CFG_FILE_EXT)

    # Create missing directories
    fileName = os.path.abspath(os.path.normpath(fileName))
    if not os.path.isdir(os.path.dirname(fileName)):
        os.makedirs(os.path.dirname(fileName))

    # If fileName already exists, make a new one by adding a number onto fileName.
    fileName = fileName.split(CFG_FILE_EXT)[0]
    cfgPath = fileName
    
    currentExt = 0
    while os.path.exists(cfgPath + CFG_FILE_EXT):
        currentExt += 1
        cfgPath = fileName + "-" + str(currentExt)

    cfgPath += CFG_FILE_EXT

    # Read default config values
    defaults = _partialModuleVariables(cfgModule)

    # Make a document to populate
    newDoc = tomlkit.document()

    # Iterate over all variables in the module
    for varName, var in defaults.items():
        # Preceeding comments are currently disabled, while variable reordering is worked around
        """
        # Does the variable have any preceeding comments?
        if retainComments and var.hasPre():
            # Add a new line, just for readability
            newDoc.add(tomlkit.nl())
            # Add all preceeding comments
            for prevDoc in var.preComments:
                newDoc.add(tomlkit.comment(prevDoc))
        """

        # Serialize the variable if required, and write it to the doc
        newDoc[varName] = _serialize(var.value, [varName], serializerKwargs=serializerKwargs)

        # Does the variable have any preceeding comments?
        if retainComments and var.hasInline():
            # Add all inline comments
            for inDoc in var.inlineComments:
                if type(newDoc[varName]) == TKContainer:
                    raise RuntimeError(f"Attempted to add a comment to a tomlkit.container.Container: {varName}")
                cast(TKItems.Item, newDoc[varName]).comment(inDoc)

    # Dump to toml and write to file
    with open(cfgPath, "w", encoding="utf-8") as f:
        f.write(tomlkit.dumps(newDoc).lstrip("\n"))

    # log and return path to new file
    log("Created " + cfgPath)
    return cfgPath


class BadTypeBehaviour(Enum):
    """Configure how to handle receiving variable values in toml that are of a different type to the corresponding
    variable in the python module.
    
    Values:
    - REJECT: Reject the value, falling back on ErrorHandling behaviour
    - KEEP: Ignore type discrepencies, and just pass the variable into the python module regardless
    - CAST: Attempt to cast the incorrectly-typed value with a call the correct type.
            E.g if the variable is of type float in python, but int is received from TOMl, this behaviour will call float(x)
            Where x is the value given in TOML
    """
    REJECT = 1
    KEEP = 2
    CAST = 3


class ErrorHandling(Enum):
    """Configure how to handle errors. Refer to the usage context of this class for the specific behaviour that will
    resut from this setting.
    
    Values:
    - RAISE: Raise an exception
    - LOG: Log the error without throwing an exception
    - IGNORE: Do nothing
    """
    RAISE = 1
    LOG = 2
    IGNORE = 3


@dataclass
class BadTypeHandling:
    """Configure how Carica should handle receiving variable values in toml that are of a different type to the corresponding
    variable in the python module, and details of logging for these events.

    The `behaviour` field sets how Carica should handle variable types in TOML that do not match the type expected by the
    python module. This allows for automatic type casting, which can be especially useful for primative types, for example:
    If the python module contains a variable that is of type `float`, but the variable is given as an `int` in toml,
    `BadTypeBehaviour.CAST` will cast the toml-given value to `float` using the `float(x)` constructor.

    The `rejectType` field sets how Carica should handle variable type rejections. If `behaviour` is `REJECT`, should
    an exception be thrown, or a log created, or neither?

    The `keepFailedCast` field is used where `behaviour` is `CAST`, setting whether to keep or reject values which fail
    casting. If an exception occurs when attempting to cast (e.g an appropriate constructor does not exist), then
    `keepFailedCast` triggers the following:
     - When `True`, keep the value from before the attempted cast, and pass this (wrongly typed) value to the python module
     - When `False`, reject this value, falling back on `rejectType` behaviour

    The `logTypeKeeping` field sets whether the keeping of variables of mismatched types should be logged. This field is used
    in two cases:
        - If `behaviour` is `KEEP`, and a value of a mismatched type is encountered
        - If `behaviour` is `CAST`, `keepFailedCast` is `True`, and a value fails type casting

    In either of these cases, if `logTypeKeeping` is `True`, a log will be created of the event. Otherwise, the value will be
    kept silently.

    The `logSuccessfulCast` field sets whether to create a log message in the event of a type-mismatched variable being casted
    to the correct type successfully. This field is only of use where `behaviour` is `CAST`.

    The `includeExceptionTrace` field sets whether exception logging should include the exception trace.

    :var behaviour: Enum setting how to handle mismatched variable types. See class for value descriptions (Default CAST)
    :vartype behaviour: BadTypeBehaviour
    :var rejectType: Enum setting the handling for the above type rejections. See class for value descriptions (Default RAISE)
    :vartype rejectType: ErrorHandling
    :var keepFailedCast: Whether to reject values that failed casting, or fall back to keeping the original (Default False)
    :vartype keepFailedCast: bool
    :var logTypeKeeping: Whether to log keeping of mismatched variable types or do so silently (Default True)
    :vartype logTypeKeeping: bool
    :var logSuccessfulCast: Whether to log successful casts of mismatched variable types (Default True)
    :vartype logSuccessfulCast: bool
    :var includeExceptionTrace: Whether to include the trace of logged exceptions, e.g in `keepFailedCast` (Default False)
    :vartype includeExceptionTrace: bool
    """
    behaviour = BadTypeBehaviour.CAST
    rejectType = ErrorHandling.RAISE
    keepFailedCast = False
    logTypeKeeping = True
    logSuccessfulCast = True
    includeExceptionTrace = False


def loadCfg(cfgModule: ModuleType, cfgFile: str, badTypeHandling: BadTypeHandling = BadTypeHandling(),
            badNameHandling: ErrorHandling = ErrorHandling.RAISE):
    """Load the values from a specified config file into attributes of the python cfg module.
    All config attributes are optional, but if a mapping is given, all desired values must be set.

    :param ModuleType cfgModule: Module to load variable values into
    :param str cfgFile: Path to the file to load. Can be relative or absolute
    :param ErrorHandling badNameHandling: How to handle variables from cfgFile which are not named in cfgModule. See class
                                            for default value and value descriptions.
    :param BadTypeHandling badTypeHandling: How to handle receiving toml variables that do not match the type of the
                                            python variable. See class for default values and value descriptions.
    :raise ValueError: When cfgFile is of an unsupported format
    """
    # Ensure the given config is toml
    if not cfgFile.endswith(CFG_FILE_EXT):
        raise ValueError("config files must be TOML")

    # Load from toml to dictionary
    with open(cfgFile, "r", encoding="utf-8") as f:
        config = tomlkit.loads(f.read())

    # Read default config values
    defaults = _partialModuleVariables(cfgModule)

    # Assign config values to cfg attributes
    for varName in config:
        varName = cast(str, varName)
        # Validate attribute names
        if varName not in defaults:
            if badNameHandling == ErrorHandling.RAISE:
                raise NameError(f"Unrecognised config variable name: {varName}")
            elif badNameHandling == ErrorHandling.LOG:
                log(f"[WARNING] Ignoring unrecognised config variable name: {varName}")

        else:
            # Get the type of the default value (or the type override) for this variable
            defaultType = defaults[varName].loadedType

            # Get the value for the variable that is defined in the config file
            # This check ignores any config attributes that do not have a value, for example comments and whitespace.
            if isinstance(config[varName], _TKItemWithValue):
                newValue: Any = cast(_TKItemWithValue, config[varName]).value
            else:
                continue

            # Convert incompatible types, e.g TOMLDocument
            if isinstance(newValue, INCOMPATIBLE_TOML_TYPES):
                newValue = convertIncompatibleTomlTypes(newValue)

            # deserialize serializable variables
            if issubclass(defaultType, SerializableType):
                newValue = defaultType.deserialize(newValue, c_badTypeHandling=badTypeHandling, c_variableTrace=[varName])

            # Handle variables of different types to that which is defined in the python module
            if not isinstance(newValue, defaultType):

                # Handle type rejections
                if badTypeHandling.behaviour == BadTypeBehaviour.REJECT:
                    errMsg = f"Unexpected type for config variable {varName}: Expected " \
                            + f"{defaultType.__name__}, received {type(newValue).__name__}"
                    if badTypeHandling.rejectType == ErrorHandling.RAISE:
                        raise TypeError(errMsg)
                    elif badTypeHandling.behaviour == ErrorHandling.LOG:
                        log(f"[WARNING] {errMsg}")

                # Handle type casts
                elif badTypeHandling.behaviour == BadTypeBehaviour.CAST:
                    if isinstance(defaultType, type):
                        try:
                            # Attempt casts for incorrect types - useful for things like ints instead of floats.
                            # TODO: Ignoring a warning here because we intentionally don't know the type or therefore the init signature
                            # In the future I will type this with a protocol before trying the instantiation
                            newValue = defaultType(newValue) # type: ignore[reportGeneralTypeIssues]
                        except Exception as e:
                            if badTypeHandling.keepFailedCast:
                                # Nothing to do if keeping failed variable casts, except log if configured to
                                if badTypeHandling.logTypeKeeping:
                                    log(f"[WARNING] Keeping original value for mistype variable {varName}, following failed " \
                                        + f"cast. Expected {defaultType.__name__}, received {type(newValue).__name__}," \
                                        + f" cast exception: {formatException(e, badTypeHandling.includeExceptionTrace)}")
                            # If configured to reject failed casts
                            else:
                                errMsg = f"Casting failed for unexpected type for config variable {varName}: Expected " \
                                        + f"{defaultType.__name__}, received {type(newValue).__name__}. " \
                                        + f"Cast exception: {formatException(e, badTypeHandling.includeExceptionTrace)}"
                                if badTypeHandling.rejectType == ErrorHandling.RAISE:
                                    raise TypeError(errMsg)
                                elif badTypeHandling.rejectType == ErrorHandling.LOG:
                                    log(f"[WARNING] {errMsg}")

                        # Cast was successful
                        else:
                            if badTypeHandling.logSuccessfulCast and type(newValue).__name__ != defaultType.__name__:
                                log(f"[WARNING] Successfully casted unexpected type for config variable {varName} from type " \
                                    + f"{type(newValue).__name__} to {defaultType.__name__}")
                    else:
                        if badTypeHandling.keepFailedCast:
                            # Nothing to do if keeping failed variable casts, except log if configured to
                            if badTypeHandling.logTypeKeeping:
                                log(f"[WARNING] Keeping original value for mistype variable {varName}, cast not attempted. " \
                                    + f"Cannot instance typing type hint {defaultType!s}. Received {type(newValue).__name__}")
                        # If configured to reject failed casts
                        else:
                            errMsg = f"Cast not attempted for unexpected type for config variable {varName}: " \
                                    + f"Cannot instance typing type hint {defaultType!s}. Received {type(newValue).__name__}"
                            if badTypeHandling.rejectType == ErrorHandling.RAISE:
                                raise TypeError(errMsg)
                            elif badTypeHandling.rejectType == ErrorHandling.LOG:
                                log(f"[WARNING] {errMsg}")


                # Handle type keeping
                elif badTypeHandling.behaviour == BadTypeBehaviour.KEEP:
                    # Nothing to do if keeping mismatched variables, except log if configured to
                    if badTypeHandling.logTypeKeeping:
                        log(f"[WARNING] Keeping original value for mistype variable {varName}. Expected " \
                            + f"{defaultType.__name__}, received {type(newValue).__name__}")

            # Variable value received successfully, inject into python module
            setattr(cfgModule, varName, newValue)

    # No errors encountered
    log("Config successfully loaded: " + cfgFile)
